import bpy, os
from bpy.types import Operator
from ..utils.utils import refresh,update_fcurves

DEBUG_MODE = False

class GRAPH_OT_Amplify(Operator):
    """Adjust the intensity of the selected F-Curves or keyframe range"""
    bl_idname = "graph.amplify"
    bl_label = "Amplify"
    bl_options = { "REGISTER", "UNDO" }

    amp_mode: bpy.props.EnumProperty(
        name= "Amplify Mode",
        default="SELECTED_FCURVE",
        description= "Amplify target to use",
        items= [("SELECTED_FCURVE","Selected F-Curves","Amplify all frames of selected F-Curves."),
                ("SELECTED_KEYS","Selected Keyframes","Amplify selected keyframes.")]
    )

    pivot_point: bpy.props.EnumProperty(
        name= "Pivot Point",
        description= "Repeating mode to use before the first keyframe",
        default="CENTER",
        items= [
            ("CENTER","Center","Amplify keys from the selection's center point."),
            ("FIRST","First","Amplify keys from the selection's first point."),
            ("MAX","Max","Amplify keys from the selection's keyframe with maximum value."),
            ("MIN","Min","Amplify keys from the selection's keyframe with minimum value."),
        ]
    )

    amount: bpy.props.FloatProperty(
        name = "Amount",
        description = "Amount by which to amplify selected F-Curves.",
        default =1.0,
        soft_min = -100, soft_max = 100)

    def get_bounding_box(self,fcurve):
        points = list()
        if self.amp_mode == "SELECTED_KEYS":
            points = [x for x in fcurve.keyframe_points if x.select_control_point == True]
            if len(points) == 0:
                return None
        elif self.pivot_point == "FIRST":
            for x in fcurve.keyframe_points:
                if x.select_control_point == True:
                    points = [x]
                    break
            if len(points) == 0:
                points = fcurve.keyframe_points
        else:
            points = fcurve.keyframe_points
        key_values = list(map(lambda x : x.co[1], points))
        maximum = max(key_values)
        minimum = min(key_values)

        if self.pivot_point == "CENTER":
            return (maximum+minimum)/2
        elif self.pivot_point == "FIRST":
            return key_values[0]
        elif self.pivot_point == "MAX":
            return maximum
        elif self.pivot_point == "MIN":
            return minimum
        return None

    def amplify(self,pivot,k):
        key_distance = pivot-k.co[1]
        hl_distance = pivot-k.handle_left[1]
        hr_distance = pivot-k.handle_right[1]
        k.co[1] = pivot-(key_distance*self.amount)
        k.handle_left[1] = pivot-(hl_distance*self.amount)
        k.handle_right[1] = pivot-(hr_distance*self.amount)

    def execute(self, context):
        scene = context.scene
        fcurves = bpy.context.selected_editable_fcurves[:]

        for fcurve in fcurves:
            pivot = self.get_bounding_box(fcurve)
            if pivot != None:
                for k in fcurve.keyframe_points:
                    if self.amp_mode == "SELECTED_KEYS":
                        if k.select_control_point:
                           self.amplify(pivot,k)
                    else:
                       self.amplify(pivot,k)

        return {'FINISHED'}

def register():
    bpy.utils.register_class(GRAPH_OT_Amplify)

def unregister():
    bpy.utils.unregister_class(GRAPH_OT_Amplify)
