import bpy, os, random, math
from bpy.types import Operator
from ..utils.utils import refresh,get_fcurve_selected_keys,update_fcurves

class GRAPH_OT_Distribute(Operator):
    """Distribute Keyframes by a specified frame amount"""
    bl_idname = "graph.distribute"
    bl_label = "Distribute"
    bl_options = { "REGISTER", "UNDO" }

    snap: bpy.props.BoolProperty(
        name= "Snap to Frame",
        default=True,
        description= "Snap keyframes to nearest frame after distribution.",
    )

    distribute_type: bpy.props.EnumProperty(
        name= "Type",
        default="SELECTED_FCURVE",
        description= "Selection to use for repeating",
        items= [("SELECTED_FCURVE","Selected F-Curves","Distribute all keyframes of selected F-Curves."),
                ("SELECTED_KEYS","Selected Keyframes","Distribute selected keyframes.")
                ]
    )

    distribute_mode: bpy.props.EnumProperty(
        name= "Mode",
        default="FRAME",
        description= "Distribute by frame or note",
        items= [("FRAME","Frames","Distribute frames by a frame number."),
                ("NOTE","Notes","Distribute frames to notes based on Scene Frame Rate.")
                ]
    )

    distribute_frames: bpy.props.FloatProperty(
        name= "Frames",
        step=100.0,
        default=2.0,
        description= "Number of frames/notes by which to distribute",
        soft_min = 1.0, soft_max = 50.0
    )

    bpm: bpy.props.FloatProperty(
        name= "BPM",
        default= 120.0,
        description= "Beats per minute",
        soft_min = 1.0, soft_max = 300.0
    )

    note_type: bpy.props.EnumProperty(
        name= "Notes",
        default="1/4",
        description= "Selection to use for repeating",
        items= [("1/1","1/1","Distribute to whole notes."),
                ("1/2","1/2","Distribute to half notes."),
                ("1/4","1/4","Distribute to quarter notes."),
                ("1/8","1/8","Distribute to eighth notes."),
                ("1/16","1/16","Distribute to sixteenth notes."),
                ("1/32","1/32","Distribute to thirth-second notes.")
                ]
    )

    def distribute(self,keyframes):
        if self.bpm <= 0:
            return

        amt = 0
        first_keyframe = 0

        if self.distribute_mode == "FRAME":
            if self.snap:
                keyframes[0].co[0] = round(keyframes[0].co[0])
            #print("first_keyframe",first_keyframe)
            first_keyframe = keyframes[0].co[0]
        elif self.distribute_mode == "NOTE":
            if self.distribute_type == "SELECTED_KEYS":
                first_keyframe = keyframes[0].co[0]
            #frames per second
            fps = bpy.context.scene.render.fps
            #frames per minute
            fpm = fps * 60
            quarter_note = fpm/self.bpm

        for i,keyframe in enumerate(keyframes):
            if self.distribute_mode == "FRAME":
                amt = first_keyframe + (i*self.distribute_frames)
            elif self.distribute_mode == "NOTE":
                if self.note_type == "1/1":
                    amt = first_keyframe + i*(quarter_note*4)
                elif self.note_type == "1/2":
                    amt = first_keyframe + i*(quarter_note*2)
                elif self.note_type == "1/4":
                    amt = first_keyframe + i*(quarter_note)
                elif self.note_type == "1/8":
                    amt = first_keyframe + i*(quarter_note*0.5)
                elif self.note_type == "1/16":
                    amt = first_keyframe + i*(quarter_note*0.25)
                elif self.note_type == "1/32":
                    amt = first_keyframe + i*(quarter_note*0.125)

            if i == 0:
                continue;

            handle_left_frame_offset = keyframe.co[0] - keyframe.handle_left[0]
            handle_left_value_offset = keyframe.co[1] - keyframe.handle_left[1]
            handle_right_frame_offset = keyframe.co[0] - keyframe.handle_right[0]
            handle_right_value_offset = keyframe.co[1] - keyframe.handle_right[1]

            if self.snap:
                keyframe.co[0] = round(amt)
                keyframe.handle_left[0] = round(amt - handle_left_frame_offset)
                keyframe.handle_right[0] = round(amt - handle_right_frame_offset)
            else:
                keyframe.co[0] = amt
                keyframe.handle_left[0] = amt - handle_left_frame_offset
                keyframe.handle_right[0] = amt - handle_right_frame_offset

    def execute(self, context):
        scene = context.scene
        fcurves = bpy.context.selected_editable_fcurves[:]
        for fcurve in fcurves:
            if self.distribute_type == "SELECTED_FCURVE":
                self.distribute(fcurve.keyframe_points)
            elif self.distribute_type == "SELECTED_KEYS":
                self.distribute(get_fcurve_selected_keys(fcurve))
        update_fcurves()
        refresh()
        return {'FINISHED'}

def register():
    bpy.utils.register_class(GRAPH_OT_Distribute)

def unregister():
    bpy.utils.unregister_class(GRAPH_OT_Distribute)
