import bpy, os
from bpy.types import Operator
from ..data_classes.animation_data import (FCurve, Keyframe)
from ..utils.utils import refresh, get_selected_keys

DEBUG_MODE = False

class GRAPH_OT_Offset(Operator):
    """Loops the selected F-Curves keyframe range"""
    bl_idname = "graph.offset"
    bl_label = "Offset"
    bl_options = { "REGISTER", "UNDO" }


    offset_mode: bpy.props.EnumProperty(
        name= "Offset Mode",
        default="SELECTED_FCURVE",
        description= "Offseting mode to use before the first keyframe",
        items= [("SELECTED_FCURVE","Selected F-Curves","Offsets all frames of selected F-Curves."),
                ("SELECTED_KEYS","Selected Keyframes","Offsets selected keyframes.")
                ]
    )

    frame_offset: bpy.props.IntProperty(
        name = "Frame Offset",
        default = 1,
        description= "Amount by which to offset the selected F-Curves.",
        soft_min = -500, soft_max = 500
    )

    keep_loop_length: bpy.props.BoolProperty(
        name= "Original Loop Length",
        default=True,
        description= "Remove keyframes added outside of original keyframe range.",
    )

    def add_key(self, fcurve, frame_number):
        frames = [ x.co[0] for x in fcurve.keyframe_points ]
        frame_index = None
        try:
            frame_index = frames.index(frame_number)
            fcurve.keyframe_points.remove(fcurve.keyframe_points[frame_index], fast=False)
            fcurve.keyframe_points.add(1)
        except ValueError:
            fcurve.keyframe_points.add(1)
        return fcurve.keyframe_points[-1]

    def copy_key(self,kf,frame,value,hl,hr):
        new_key = Keyframe([frame,value],hl,hr)
        new_key.interpolation = kf.interpolation
        new_key.amplitude = kf.amplitude
        new_key.back = kf.back
        new_key.easing = kf.easing
        new_key.period = kf.period
        new_key.type = kf.type
        new_key.handle_left_type = kf.handle_left_type
        new_key.handle_right_type = kf.handle_right_type
        return new_key

    def find_FCurve(self,data_path,array_index):
        fcurves = bpy.context.selected_editable_fcurves[:]
        for fcurve in fcurves:
            if fcurve.data_path == data_path and fcurve.array_index == array_index:
                return fcurve

    def offset_fcurve(self,fcurve,loop_length):
        amount = self.frame_offset

        if abs(self.frame_offset) > loop_length:
            difference = self.frame_offset-loop_length
            #print("difference",difference,"loop_length",loop_length,"mod",self.frame_offset % loop_length)
            amount = self.frame_offset % loop_length

        for keyframe in fcurve.keyframe_points:
            keyframe.co[0] += amount
            keyframe.handle_left[0] += amount
            keyframe.handle_right[0] += amount

    def build_loops(self,fcurve):
        counts = [1,1]
        new_fcurve = FCurve(fcurve.data_path,fcurve.array_index,list())
        for i,count in enumerate(counts):
            selection = get_selected_keys(fcurve,self.offset_mode)

            start_frame = selection[0].co[0]
            #start_value = selection[0].co[1]
            end_frame = selection[-1].co[0]
            #end_value = selection[-1].co[1]

            for loop_index in range(1,count+1):
                key_range = end_frame - start_frame
                frame_shift = (key_range * loop_index)
                #loop_length = frame_shift + start_frame + end_frame
                value_shift = 0

                BEFORE_MODE = False
                AFTER_MODE = False

                if i == 0:
                    BEFORE_MODE = True
                elif i == 1:
                    AFTER_MODE = True

                loop = list()
                if BEFORE_MODE: #before_count
                    frame_shift *= -1
                    value_shift *= -1

                for kf in selection:
                    frame_number = 0
                    frame_number = kf.co[0] + frame_shift

                    l_handle = kf.handle_left[0]+frame_shift
                    r_handle = kf.handle_right[0]+frame_shift

                    l_value = kf.handle_left[1]+value_shift
                    r_value = kf.handle_right[1]+value_shift

                    frame_value = kf.co[1]+value_shift

                    new_key = self.copy_key(kf,
                        frame_number,
                        frame_value,
                        [l_handle,l_value],
                        [r_handle,r_value],
                    )

                    loop.append(new_key)

                if BEFORE_MODE: #before_count
                    loop.reverse()
                new_fcurve.keys.append(loop)
        return new_fcurve

    def find_keyframe_by_frame(self,fcurve,frame):
        for keyframe in fcurve.keyframe_points:
            if keyframe.co[0] == frame:
                return keyframe

    def execute(self, context):
        fcurves = bpy.context.selected_editable_fcurves[:]
        new_fcurves = []

        for fi,fcurve in enumerate(fcurves):
            new_fcurve = self.build_loops(fcurve)
            new_fcurves.append(new_fcurve)

        for j,fcurve in enumerate(fcurves):
            first_keyframe = fcurve.keyframe_points[0].co[0]
            last_keyframe = fcurve.keyframe_points[-1].co[0]
            loop_length = last_keyframe-first_keyframe

            for loop in new_fcurves[j].keys:
                for keyframe in loop:
                    new_key = fcurve.keyframe_points.insert(keyframe.co[0], keyframe.co[1], options={"FAST"})
                    new_key.interpolation = keyframe.interpolation
                    new_key.amplitude = keyframe.amplitude
                    new_key.back = keyframe.back
                    new_key.easing = keyframe.easing
                    new_key.period = keyframe.period
                    new_key.type = keyframe.type
                    new_key.handle_left_type = keyframe.handle_left_type
                    new_key.handle_right_type = keyframe.handle_right_type
                    new_key.handle_left[0] = keyframe.handle_left[0]
                    new_key.handle_left[1] = keyframe.handle_left[1]
                    new_key.handle_right[0] = keyframe.handle_right[0]
                    new_key.handle_right[1] = keyframe.handle_right[1]
           
            self.offset_fcurve(fcurve,loop_length)            

            if (self.keep_loop_length):
                frames = [ keyframe.co[0] for keyframe in fcurve.keyframe_points ]

                if first_keyframe not in frames:
                    fcurve.keyframe_points.insert(first_keyframe, fcurve.evaluate(first_keyframe), options={"FAST"})

                if last_keyframe not in frames:
                    fcurve.keyframe_points.insert(last_keyframe, fcurve.evaluate(last_keyframe), options={"FAST"})

                frames = [ keyframe.co[0] for keyframe in fcurve.keyframe_points ]

                reversed_keys = fcurve.keyframe_points[:]
                reversed_keys.reverse()

                for i,keyframe in enumerate(reversed_keys):
                    try:
                        if keyframe.co[0] < first_keyframe:
                            fcurve.keyframe_points.remove(keyframe, fast=True)
                    except ReferenceError:
                        pass

                    try:
                        if keyframe.co[0] > last_keyframe:
                            fcurve.keyframe_points.remove(keyframe, fast=True)
                    except ReferenceError:
                        pass

            fcurve.update()

        refresh()
        return {'FINISHED'}

def register():
    bpy.utils.register_class(GRAPH_OT_Offset)

def unregister():
    bpy.utils.unregister_class(GRAPH_OT_Offset)