import bpy
from bpy.types import Operator
from ..utils.utils import (refresh, loop_keys, get_selected_keys)

class GRAPH_OT_PerfectLoop(Operator):
    """Creates seamless loops for the selected F-Curves"""
    bl_idname = "graph.perfect_loop"
    bl_label = "Perfect Loop"
    bl_options = { "REGISTER", "UNDO" }

    loop_mode: bpy.props.EnumProperty(
        name= "Loop Mode",
        default="SELECTED_FCURVE",
        description= "Mode for creating the perfect loop with options "\
            "for entire F-Curve or selected keyframes",
        items= [("SELECTED_FCURVE","Selected F-Curves","Loops all frames of selected F-Curves."),
                ("SELECTED_KEYS","Selected Keyframes","Loops selected keyframes.")
                ]
    )

    vector: bpy.props.BoolProperty(
        name = "Vectorize Start & End",
        description = "Aligns start and end keyframes' handles to vector handles.",
        default = False
    )

    match_mode: bpy.props.EnumProperty(
        name= "Match",
        default="MATCH_FIRST",
        description= "Keyframe to use for creating a loop",
        items= [("MATCH_FIRST","First","Uses the F-Curve's first keyframe's values to create a loop."),
                ("MATCH_LAST","Last","Uses the F-Curve's last keyframe's values to create a loop.")
                ]
    )

    match_angles: bpy.props.BoolProperty(
        name= "Match Angles",
        default=True,
        description= "Adjust start and end keyframes' keyframe angles for smooth transition.",
    )

    match_lengths: bpy.props.BoolProperty(
        name= "Match Lengths",
        default=False,
        description= "Match start and end keyframes' keyframe lengths.",
    )

    def get_fcurve_selected_keys(self, fcurve, no_limits=False):
        selected_keys = []

        for keyframe in fcurve.keyframe_points:
            if keyframe.select_control_point:
                selected_keys.append(keyframe)

        if no_limits == True:
            selected_keys = selected_keys[1:-1]

        return selected_keys

    def convert_to_vector(self, keyframe):
        keyframe.handle_left_type = "VECTOR"
        keyframe.handle_right_type = "VECTOR"

    def convert_to_aligned(self, keyframe):
        keyframe.handle_left_type = "ALIGNED"
        keyframe.handle_right_type = "ALIGNED"

    def convert_to_previous(self, keyframe, previous):
        keyframe.handle_left_type = previous[0]
        keyframe.handle_right_type = previous[1]

    def get_handles(self, keyframe):
        return [
            str(keyframe.handle_left_type),
            str(keyframe.handle_right_type)
        ]

    def execute(self, context):
        fcurves = bpy.context.selected_editable_fcurves[:]

        for i,fcurve in enumerate(fcurves):
            selected_keys = get_selected_keys(fcurve,self.loop_mode)
            first_key = selected_keys[0]
            last_key = selected_keys[-1]
            #first_key_handles = self.get_handles(first_key)
            #last_key_handles = self.get_handles(last_key)
            if self.vector == True:
                self.convert_to_vector(first_key)
                self.convert_to_vector(last_key)
                fcurve.update()
                self.convert_to_aligned(first_key)
                self.convert_to_aligned(last_key)
                fcurve.update()
                """
                self.convert_to_previous(first_key,first_key_handles)
                self.convert_to_previous(last_key,last_key_handles)
                fcurve.update()
                """
            else:
                fcurve.update()                

            loop_keys(self.match_mode,self.match_angles,self.match_lengths,first_key,last_key)

        refresh()
        return {'FINISHED'}

def register():
    bpy.utils.register_class(GRAPH_OT_PerfectLoop)

def unregister():
    bpy.utils.unregister_class(GRAPH_OT_PerfectLoop)