import bpy, os, random
from bpy.types import Operator
from ..utils.utils import get_selected_keys,convert_to_auto
class GRAPH_OT_Randomize(Operator):
    """Randomize selected keyframes' frames or values by a specified amount"""
    bl_idname = "graph.randomize"
    bl_label = "Randomize"
    bl_options = { "REGISTER", "UNDO" }

    randomize_mode: bpy.props.EnumProperty(
        name= "Randomize Mode",
        default="SELECTED_FCURVE",
        description= "Mode for assigning random values to keyframes.",
        items= [("SELECTED_FCURVE","Selected F-Curves","Randomizes all keyframes of selected F-Curves."),
                ("SELECTED_KEYS","Selected Keyframes","Randomizes selected keyframes.")
                ]
    )

    axis: bpy.props.EnumProperty(
        name= "Axis",
        default="Y_AXIS",
        description= "Randomize method",
        items= [
            ("X_AXIS","X","Randomize keyframes by frames."),
            ("Y_AXIS","Y","Randomize keyframes by values."),
            ("BOTH","Both","Randomize keyframes by both frames and values."),
        ]
    )
    absolute: bpy.props.BoolProperty(
        name = "Absolute",
        description = "Use absolute values.",
        default = False
        )

    snap: bpy.props.BoolProperty(
        name = "Snap to Frame",
        description = "Snap randomized frames to whole frame values.",
        default = True
     )

    keep_sign: bpy.props.BoolProperty(
        name = "Keep Sign",
        description = "Prevent keyframe values from changing from positive"\
        "to negative or vice versa.",
        default = True
        )

    min: bpy.props.FloatProperty(
        name = "Min",
        description = "Minimum amount for randomized keyframes.",
        default = -10.0,
        soft_min = -1000, soft_max = 1000)

    max: bpy.props.FloatProperty(
        name = "Max",
        description = "Maximum amount for randomized keyframes.",
        default = 10.0,
        soft_min = -1000, soft_max = 1000)

    amount: bpy.props.FloatProperty(
        name = "Amount",
        description = "Amount of randomization to apply.",
        default = 100.0,
        soft_min = 0, soft_max = 300)

    random_seed: bpy.props.IntProperty(
        name = "Random Seed",
        description = "Random seed for random amounts.",
        default=0,
        soft_min = 0, soft_max = 1000)

    exclude_start_and_end: bpy.props.BoolProperty(
        name = "Exclude Start & End",
        description = "Excludes the start & end frame from randomization "\
            "in the case of looping F-Curve animations.",
        default = False
     )

    automatic: bpy.props.BoolProperty(
        name = "Convert to Automatic",
        description = "Converts selected keyframes' handles to automatic handles.",
        default = False
     )

    def get_curve_value_limits(self,fcurve):
        values = [ keyframe.co[1] for keyframe in fcurve.keyframe_points ]
        return min(values),max(values)

    def get_curve_frame_limits(self,fcurve):
        frames = [ keyframe.co[0] for keyframe in fcurve.keyframe_points ]
        return min(frames),max(frames)

    def execute(self, context):
        rotation_dps = ["rotation_euler","rotation_quaternion",
        "delta_rotation_quaternion","delta_rotation_euler"]

        fcurves = bpy.context.selected_editable_fcurves[:]
        for fcurve in fcurves:
            selected_keys = get_selected_keys(fcurve,self.randomize_mode,
                exclude_limits=self.exclude_start_and_end)

            #min_value,max_value = self.get_curve_value_limits(fcurve)
            #min_frame,max_frame = self.get_curve_frame_limits(fcurve)

            #print("min_value: %s, max_value: %s"%(min_value,max_value))
            #print("min_frame: %s, max_frame: %s"%(min_frame,max_frame))

            for i, sk in enumerate(selected_keys):
                random.seed(self.random_seed*len(selected_keys) + i)
                amt = self.amount/100
                if fcurve.data_path in rotation_dps:
                    amt = self.amount/10000
                #val = random.uniform(min_value*-amt, max_value*amt)
                val = random.uniform(self.min, self.max) * amt
                #frm = random.uniform(min_frame*-amt, max_frame*amt)

                if self.axis == "X_AXIS" or self.axis == "BOTH":
                    if self.absolute:
                        sk.co[0] = val
                        sk.handle_left[0] = val
                        sk.handle_right[0] = val
                    else:
                        sk.co[0] += val
                        sk.handle_left[0] += val
                        sk.handle_right[0] += val
                    if self.snap == True:
                        sk.co[0] = round(sk.co[0])
                        sk.handle_left[0] = round(sk.handle_left[0])
                        sk.handle_right[0] = round(sk.handle_right[0])

                if self.axis == "Y_AXIS" or self.axis == "BOTH":
                    if (self.keep_sign):
                        sign = lambda a: (a>0) - (a<0)
                        key_sign = sign(sk.co[1])
                        sign_val = sign(val)
                        if key_sign != sign_val:
                            val *= -1

                    if self.absolute:
                        sk.co[1] = val
                        sk.handle_left[1] = val
                        sk.handle_right[1] = val
                    else:
                        sk.co[1] += val
                        sk.handle_left[1] += val
                        sk.handle_right[1] += val

        if self.axis == "X_AXIS" or self.axis == "BOTH":
            bpy.ops.graph.clean(channels=False)

        if self.automatic == True:
            convert_to_auto(fcurves,self.exclude_start_and_end)

        return {'FINISHED'}

def register():
    bpy.utils.register_class(GRAPH_OT_Randomize)

def unregister():
    bpy.utils.unregister_class(GRAPH_OT_Randomize)
