import bpy
from bpy.types import Operator
from bpy.props import IntProperty, BoolProperty
from ..utils.utils import refresh,most_frequent

class GRAPH_OT_RangeToLoop(Operator):
    """Sets frame range based on loop length"""
    bl_idname = "graph.range_to_loop"
    bl_label = "Range to Loop"
    bl_options = { "REGISTER", "UNDO" }

    start_frame: IntProperty(
        name= "Start Frame",
        default= 1,
        description= "Start frame for looping animation",
        soft_min = 0, soft_max = 5000
    )

    end_frame: IntProperty(
        name= "End Frame",
        default= 1,
        options={'HIDDEN'},
        description= "End frame for looping animation",
    )

    loop_length: IntProperty(
        name= "Loop Length",
        default= 24,
        description= "Loop length for setting frame range",
        soft_min = 1, soft_max = 500,
    )

    loop_count: IntProperty(
        name= "Loop Count",
        default= 2,
        description= "Loop count for setting frame range",
        soft_min = 1, soft_max = 500,
    )

    preview_range: BoolProperty(
        name= "Preview Range",
        default= False,
        description= "Set Preview Frame Range to Loop"
    )

    scene_range: BoolProperty(
        name= "Scene Range",
        default= True,
        description = "Sets Scene Frame Range to Loop"
    )
    
    def get_loop(self):
        fcurves = bpy.context.selected_editable_fcurves[:]
        if len(fcurves) == 0:
            fcurves = bpy.context.visible_fcurves[:]
            if len(fcurves) == 0:
                return None,None

        first_frames = list()
        last_frames = list()

        for fcurve in fcurves:
            first_keyframe = fcurve.keyframe_points[0]
            fcurve_length = len(fcurve.keyframe_points)-1
            last_keyframe = fcurve.keyframe_points[fcurve_length]
            if first_keyframe.co[0] != last_keyframe.co[0]:
                first_frames.append(int(round(first_keyframe.co[0])))
                last_frames.append(int(round(last_keyframe.co[0])))

        most_common_first = most_frequent(first_frames)
        most_common_last = most_frequent(last_frames)
        loop_length = most_common_last-most_common_first

        for fcurve in fcurves:
            for keyframe in fcurve.keyframe_points:
                if keyframe.co[0] < most_common_first:
                    keyframe.select_control_point = False
                elif keyframe.co[0] > most_common_last:
                    keyframe.select_control_point = False

        return most_common_first, most_common_last, loop_length

    def invoke(self, context, event):
        most_common_first, most_common_last, loop_length = self.get_loop()
        if most_common_first and loop_length:
            self.start_frame = most_common_first
            self.end_frame = most_common_last
            self.loop_length = loop_length
        
        return self.execute(context)

    def execute(self, context):
        fcurves = bpy.context.selected_editable_fcurves[:]
        fcurve = fcurves[0]
        if self.preview_range == True:
            bpy.data.scenes["Scene"].frame_preview_start = self.start_frame
            bpy.data.scenes["Scene"].frame_preview_end = self.start_frame + (self.loop_length*self.loop_count) - 1
            bpy.data.scenes["Scene"].use_preview_range = True
        if self.scene_range == True:
            bpy.data.scenes["Scene"].frame_start = self.start_frame
            bpy.data.scenes["Scene"].frame_end = self.start_frame + (self.loop_length*self.loop_count) - 1
        return {'FINISHED'}


def register():
    bpy.utils.register_class(GRAPH_OT_RangeToLoop)

def unregister():
    bpy.utils.unregister_class(GRAPH_OT_RangeToLoop)