import bpy, os
from bpy.types import Operator
from ..data_classes.animation_data import (FCurve, Keyframe)
from ..utils.utils import (refresh, get_selected_keys)

DEBUG_MODE = False
class GRAPH_OT_Repeat(Operator):
    """Loops the selected F-Curves keyframe range"""
    bl_idname = "graph.repeat"
    bl_label = "Repeat"
    bl_options = { "REGISTER", "UNDO" }


    repeat_mode: bpy.props.EnumProperty(
        name= "Repeat Mode",
        default="SELECTED_FCURVE",
        description= "Repeating mode to use before the first keyframe",
        items= [("SELECTED_FCURVE","Selected F-Curves","Repeats all frames of selected F-Curves."),
                ("SELECTED_KEYS","Selected Keyframes","Repeats selected keyframes.")
                ]
    )

    before_mode: bpy.props.EnumProperty(
        name= "Before Mode",
        default="REPEAT_MOTION",
        description= "Repeating mode to use before the first keyframe",
        items= [("REPEAT_MOTION","Repeat Motion","Repeats keyframe range as-is."),
                ("REPEAT_OFFSET","Repeat with Offset","Repeats keyframe range "\
                "with continued rate of change from start & end keyframe."),
                ("REPEAT_MIRROR","Repeat Mirrored","Alternate between forward and "\
                "reverse playback of keyframe range."),
            ]
    )

    before_count: bpy.props.IntProperty(
        name = "Count",
        description = "Number of times to repeat the keyframe range",
        default = 0,
        soft_min = 0, soft_max = 100
    )

    after_mode: bpy.props.EnumProperty(
        name= "After Mode",
        default="REPEAT_MOTION",
        description= "Repeating mode to use before the first keyframe",
        items= [("REPEAT_MOTION","Repeat Motion","Repeats keyframe range as-is."),
                ("REPEAT_OFFSET","Repeat with Offset","Repeats keyframe range "\
                "with continued rate of change from start & end keyframe."),
                ("REPEAT_MIRROR","Repeat Mirrored","Alternate between forward and "\
                "reverse playback of keyframe range."),
            ]
    )

    after_count: bpy.props.IntProperty(
        name = "Count",
        description = "Number of times to repeat the keyframe range",
        default = 1,
        soft_min = 0, soft_max = 100
    )

    def add_key(self, fcurve, frame_number):
        frames = [ x.co[0] for x in fcurve.keyframe_points ]
        frame_index = None
        try:
            frame_index = frames.index(frame_number)
            fcurve.keyframe_points.remove(fcurve.keyframe_points[frame_index], fast=False)
            fcurve.keyframe_points.add(1)
        except ValueError:
            fcurve.keyframe_points.add(1)
        return fcurve.keyframe_points[-1]

    def copy_key(self,kf,frame,value,hl,hr):
        new_key = Keyframe([frame,value],hl,hr)
        new_key.interpolation = kf.interpolation
        new_key.amplitude = kf.amplitude
        new_key.back = kf.back
        new_key.easing = kf.easing
        new_key.period = kf.period
        new_key.type = kf.type
        new_key.handle_left_type = kf.handle_left_type
        new_key.handle_right_type = kf.handle_right_type
        return new_key

    def find_FCurve(self,data_path,array_index):
        fcurves = bpy.context.selected_editable_fcurves[:]
        for fcurve in fcurves:
            if fcurve.data_path == data_path and fcurve.array_index == array_index:
                return fcurve

    def convert_limits_to_vector(self, fcurves):
        for fcurve in fcurves:
            selected_keys = get_selected_keys(fcurve,self.repeat_mode)
            first = selected_keys[0]
            last = selected_keys[-1]
            first.handle_left_type = "VECTOR"
            first.handle_right_type = "VECTOR"
            last.handle_left_type = "VECTOR"
            last.handle_right_type = "VECTOR"
            fcurve.update()

    def build_loops(self,fcurve):
        counts = [self.before_count,self.after_count]
        new_fcurve = FCurve(fcurve.data_path,fcurve.array_index,list())
        for i,count in enumerate(counts):
            selection = get_selected_keys(fcurve,self.repeat_mode)

            start_frame = selection[0].co[0]
            start_value = selection[0].co[1]
            end_frame = selection[-1].co[0]
            end_value = selection[-1].co[1]

            for loop_index in range(1,count+1):
                key_range = end_frame - start_frame
                frame_offset = key_range * loop_index
                loop_length = frame_offset + start_frame + end_frame
                value_offset = 0

                BEFORE_MODE = False
                AFTER_MODE = False

                if i == 0:
                    BEFORE_MODE = True
                elif i == 1:
                    AFTER_MODE = True

                loop = list()
                if BEFORE_MODE: #before_count
                    frame_offset *= -1
                    value_offset *= -1
                    if self.before_mode == "REPEAT_OFFSET":
                        value_offset = (end_value-start_value)*loop_index
                elif AFTER_MODE:
                    if self.after_mode == "REPEAT_OFFSET":
                        value_offset = (end_value-start_value)*loop_index

                for ki,kf in enumerate(selection):
                    frame_number = 0
                    if AFTER_MODE and (self.after_mode == "REPEAT_MIRROR" and (loop_index % 2) != 0):
                        #after loops
                        frame_number =  loop_length + kf.co[0]*-1

                        if ki == len(selection)-1:
                            left_handle_offset = selection[-1].co[0]-selection[-1].handle_left[0]
                            l_handle = frame_number-left_handle_offset
                            r_handle = loop_length + kf.handle_left[0] * -1
                        else:
                            l_handle = loop_length + kf.handle_right[0] * -1
                            r_handle = loop_length + kf.handle_left[0] * -1

                        l_value = kf.handle_right[1]
                        r_value = kf.handle_left[1]
                    elif BEFORE_MODE and (self.before_mode == "REPEAT_MIRROR" and (loop_index % 2) != 0):
                        #before loops
                        offset = frame_offset + start_frame + end_frame
                        frame_number = offset + kf.co[0]*-1

                        if ki == 0:
                            end_handle = selection[0].handle_right[0]-selection[0].co[0]
                            l_handle = offset + kf.handle_right[0] * -1
                            r_handle = frame_number + end_handle
                        else:
                            l_handle = offset + kf.handle_right[0] * -1
                            r_handle = offset + kf.handle_left[0] * -1

                        l_value = kf.handle_right[1]
                        r_value = kf.handle_left[1]
                    else:
                        frame_number = kf.co[0]+frame_offset

                        if (BEFORE_MODE and self.before_mode == "REPEAT_MIRROR") or \
                        (AFTER_MODE and self.after_mode == "REPEAT_MIRROR"):
                            if ki == 0:
                                left_handle_offset = selection[0].co[0]-selection[0].handle_left[0]
                                right_handle_offset = selection[0].handle_right[0]-selection[0].co[0]
                                l_handle = frame_number-right_handle_offset
                                r_handle = kf.handle_right[0]+frame_offset
                            else:
                                l_handle = kf.handle_left[0]+frame_offset
                                r_handle = kf.handle_right[0]+frame_offset
                        else:
                            l_handle = kf.handle_left[0]+frame_offset
                            r_handle = kf.handle_right[0]+frame_offset

                        l_value = kf.handle_left[1]+value_offset
                        r_value = kf.handle_right[1]+value_offset

                    frame_value = kf.co[1]+value_offset
                    #if DEBUG_MODE: print("frame_value",frame_value)

                    new_key = self.copy_key(kf,
                        frame_number,
                        frame_value,
                        [l_handle,l_value],
                        [r_handle,r_value],
                    )

                    if BEFORE_MODE and (self.before_mode == "REPEAT_MIRROR" and (loop_index % 2) != 0) or \
                    AFTER_MODE and (self.after_mode == "REPEAT_MIRROR" and (loop_index % 2) != 0):
                        loop.insert(0, new_key)
                    else:
                        loop.append(new_key)

                if BEFORE_MODE: #before_count
                    loop.reverse()
                new_fcurve.keys.append(loop)
        return new_fcurve

    def execute(self, context):
        fcurves = bpy.context.selected_editable_fcurves[:]
        new_fcurves = []

        for fi,fcurve in enumerate(fcurves):
            new_fcurve = self.build_loops(fcurve)
            new_fcurves.append(new_fcurve)

        for j,fcurve in enumerate(fcurves):
            if self.before_count > 0:
                fcurve.keyframe_points.remove(fcurve.keyframe_points[0], fast=True)
            if self.after_count > 0:
                fcurve.keyframe_points.remove(fcurve.keyframe_points[-1], fast=True)

            for loop in new_fcurves[j].keys:
                for keyframe in loop:
                    new_key = fcurve.keyframe_points.insert(keyframe.co[0], keyframe.co[1], options={"FAST"})
                    new_key.interpolation = keyframe.interpolation
                    new_key.amplitude = keyframe.amplitude
                    new_key.back = keyframe.back
                    new_key.easing = keyframe.easing
                    new_key.period = keyframe.period
                    new_key.type = keyframe.type
                    new_key.handle_left_type = keyframe.handle_left_type
                    new_key.handle_right_type = keyframe.handle_right_type
                    new_key.handle_left[0] = keyframe.handle_left[0]
                    new_key.handle_left[1] = keyframe.handle_left[1]
                    new_key.handle_right[0] = keyframe.handle_right[0]
                    new_key.handle_right[1] = keyframe.handle_right[1]

        refresh()
        return {'FINISHED'}

def register():
    bpy.utils.register_class(GRAPH_OT_Repeat)

def unregister():
    bpy.utils.unregister_class(GRAPH_OT_Repeat)